/***************************************************
 *** Google Talk Bot library
 *** Ross Andrews
 *** January 2008
 *** 
 *** To use, add smack.jar and smackx.jar to the CP,
 *** load this file, and call gtBot().
 ***************************************************/

//////////////////////////////////////////////////
/// Imports
//////////////////////////////////////////////////

/*extern importClass Packages JavaAdapter print Continuation */
importClass(Packages.java.lang.Thread); /*extern Thread */ /* This is just handy to have around */
importClass(Packages.org.jivesoftware.smack.ConnectionConfiguration); /*extern ConnectionConfiguration */
importClass(Packages.org.jivesoftware.smack.XMPPConnection); /*extern XMPPConnection */
importClass(Packages.org.jivesoftware.smack.MessageListener); /*extern MessageListener */
importClass(Packages.org.jivesoftware.smack.PacketListener); /*extern PacketListener */
importClass(Packages.org.jivesoftware.smack.Roster); /*extern Roster */
importClass(Packages.org.jivesoftware.smack.filter.PacketFilter); /*extern PacketFilter */
importClass(Packages.org.jivesoftware.smack.filter.MessageTypeFilter); /*extern MessageTypeFilter */
importClass(Packages.org.jivesoftware.smack.filter.PacketTypeFilter); /*extern PacketTypeFilter */
importClass(Packages.org.jivesoftware.smack.filter.AndFilter); /*extern AndFilter */
importClass(Packages.org.jivesoftware.smack.filter.FromMatchesFilter); /*extern FromMatchesFilter */
importClass(Packages.org.jivesoftware.smack.packet.Presence); /*extern Presence */
importClass(Packages.org.jivesoftware.smack.packet.Message); /*extern Message */

//////////////////////////////////////////////////
/// Constants
//////////////////////////////////////////////////

var HOST='talk.google.com';
var PORT=5222;
var SERVICE='gmail.com';

//////////////////////////////////////////////////
/// Lower-level utility functions
//////////////////////////////////////////////////

function gt_setPresence(conn,avail){
    var pres=new Presence(Presence.Type.available);
    if(avail){
	pres.setMode(Presence.Mode.available);
    }else{
	pres.setMode(Presence.Mode.away);
    }
    conn.sendPacket(pres);
}

function gt_login(username,password){
    var connConfig = new ConnectionConfiguration(HOST, PORT, SERVICE);
    var xc=new XMPPConnection(connConfig);
    xc.connect();
    xc.login(username,password);

    gt_setPresence(xc,true);

    // This is needed to let people talk who aren't on the buddy list
    xc.getRoster().setSubscriptionMode(Roster.SubscriptionMode.accept_all);
    return xc;
}

function gt_logout(conn){
    var pres=new Presence(Presence.Type.unavailable);
    conn.sendPacket(pres);
    conn.disconnect();
}

function gt_sendMessage(conn,to,message){
    if(message===null || message===''){return;} // don't send empty messages
    var msg=new Message(to);
    msg.setBody(message);
    conn.sendPacket(msg); // send it
}

// Adds a handler for all messages from a certain source.
function gt_addChatter(wrapper,from,chatter){

    var processPacket=function(packet,wrapper){
	if(packet.getBody()!==null){ // null bodies come through for "typing" messages
	    // The ''+ is to convert it to a JS string, so that === works like you'd expect.
	    var body=chatter(''+packet.getBody(),wrapper); // Get our response
	    gt_sendMessage(wrapper.conn,packet.getFrom(),body);
	}
    };

    var packetListener=new JavaAdapter(PacketListener,{
	    processPacket:function(packet){
		processPacket(packet,wrapper);
	    }
	});

    var packetFilter=new AndFilter();
    packetFilter.addFilter(new PacketTypeFilter(Message));
    packetFilter.addFilter(new FromMatchesFilter(from));
    
    wrapper.conn.addPacketListener(packetListener,packetFilter);

    // We return this because you might want to immediately 
    // call it yourself, with some packet you got that 
    // prompted you to make a Chatter.
    return processPacket;
}

// Adds a handler for all messages, ignoring other packet types.
function gt_addGlobalHandler(conn,handler){
    var packetListener=new JavaAdapter(PacketListener,{
	    processPacket:handler
	});

    var packetFilter=new PacketTypeFilter(Message);
    conn.addPacketListener(packetListener,packetFilter);
}

//////////////////////////////////////////////////
/// Connection wrapper
//////////////////////////////////////////////////

function ConnectionWrapper(conn){
    this.conn=conn;
    return this;
}

ConnectionWrapper.prototype={
    sendMessage:function(to,msg){
	gt_sendMessage(this.conn,to,msg);
    }
};

//////////////////////////////////////////////////
/// The actual framework
//////////////////////////////////////////////////
function gtBot(username,password,chatterFactory){
    var conn=gt_login(username,password);
    var clients={}; // name->handler fn
    var jid=/^(.*?)\//;
    var wrapper=new ConnectionWrapper(conn);

    gt_addGlobalHandler(conn,function(packet){
	    var from=packet.getFrom();

	    if(from!==null && clients[from]===undefined){
		print('New client '+from+', making new chatter.');
		clients[from]=chatterFactory(from);
		var handler=gt_addChatter(wrapper,from,clients[from]);
		
		if(packet.getBody()!==null){
		    handler(packet,wrapper); // Call it the first time
		}
	    } else {
		//print(packet.toXML());
	    }
	});

    return function(){
	gt_logout(conn);
    };
}

//////////////////////////////////////////////////
/// Continuation bots
//////////////////////////////////////////////////
var gt_sessions={}; // List of continuations, by username

/*
 * Wraps a responder (username,suspend)=>(message,wrapper) 
 * in a function that will clear out the session after it returns.
 */
function gt_wrapResponder(fn){
    return function(username, suspend){
	var unwrapped=fn(username, suspend);
	return function(msg,connWrapper){ // this gets passed to gtBot
	    var finalMsg=unwrapped(msg,connWrapper); // when we finally return to here, it's the final message
	    gt_sessions[username]=null; // clear the session
	    return finalMsg; // give the last word to gtBot
	};
    };
}


function gtContinuationBot(user,password,responder){
    var logout=gtBot(user,password,function(username){
	    
	    return function(msg,connWrapper){
		var respond=new Continuation(); // The caller here is where we will send intermediate responses to.
		
		if(gt_sessions[username]){ // session exists, call it
		    gt_sessions[username](msg);
		}else{ // No session yet, make one
		    var suspender=function(response){ // This is what sends intermediate responses
			gt_sessions[username]=new Continuation(); // Return back to the wrapped responder
			respond(response); // Send the response back to gtBot
		    };
		    
		    var wrapped=gt_wrapResponder(responder); // Makes a wrapped-responder-maker
		    var wrappedResponder=wrapped(username,suspender); // this has the same signature as responder
		    return wrappedResponder(msg,connWrapper); // And kick things off with the first message!
		}
	    };
	});
    return logout;
}


//////////////////////////////////////////////////
/// Test cases
//////////////////////////////////////////////////
function gt_test_parrotFactory(username){
    print('I am a chatter for '+username);
    return function(msg,wrapper){
	print(username+' said '+msg);

	if(msg=='toggle'){
	    wrapper.something();
	    return null;
	}else{
	    return 'You said "'+msg+'".';
	}
    };
}

function gt_test_wrapper(username){
    return function(msg,wrapper){
	wrapper.sendMessage(msg,username+" says hey!");
	return null;
    };
}

function gt_test_continuation(username,suspend){
    return function(message){
	var last=message;
	while(last!='exit'){
	    last=suspend("You said '"+last+"'. Say 'exit' to end the session.");
	}
	return 'Ending the session...';
    };
}

function gt_test_memory(from,suspend){
    return function(message,wrapper){
	wrapper.sendMessage(from,"Type 'help' for help.");
	var memory={};
	var keys=[];

	while(message != 'exit'){
	    var match=null;
	    var key=null;

	    if(message=='help'){
		message=suspend("This app remembers things as long as you use it.\n"+
				"\t'put: <key> <value>' to store something\n"+
				"\t'get: <key>' to retrieve it\n"+
				"\t'list' to show all keys\n"+
				"\t'exit' to quit, clearing the memory");
	    }else if(message=='list'){
		var list='';
		for(var k=0;k<keys.length;k++){
		    if(list!==''){list+=', ';}
		    list+=keys[k];
		}
		message=suspend('Keys: '+(list===''?'<none>':list));
	    }else if((match=message.match(/^put:\s*(\S+)\s*(.*)/))){
		key=match[1];
		var value=match[2];
		if(memory[key]===undefined){
		    keys.push(key);
		}
		memory[key]=value;
		message=suspend("Stored in "+key+". Total of "+keys.length+" thing"+(keys.length==1?"":"s")+" remembered.");
	    }else if((match=message.match(/^get:\s*(\S*)/))){
		    key=match[1];
		    var msg=memory[key];
		    if(msg===null || msg===undefined || msg===''){
			msg="<empty>";
		    }
		    message=suspend(msg);
	    }else{
		message=suspend("I don't understand '"+message+"'.");
	    }
	}
	return 'Exiting...';
    };
}