/*extern Packages importClass */
importClass(Packages.java.util.Random); /*extern Random */

var rand=new Random();

function ntimes(n,lambda){
    for(var i=0;i<n;i++){
	lambda(i);
    }
}

// Can be called with one or two args.
// If two, then rolls dice.
// If one, then returns the one, as a number.
function roll(num,sides){
    if(sides===undefined){ // only called with one arg, so it's a constant
	return Number(num);
    }else{
	var sum=0;
	ntimes(num,function(){
		sum+=rand.nextInt(sides)+1;
	    });
	return sum;
    }
}

// takes "4d20" and returns [4,20]
// or '43' and returns [43]
// null if not in right form
var clauseRx=/^([0-9]+)d([0-9]+)$/;
var constRx=/^[0-9]+$/;
function parseClause(clause){
    var m=clause.match(clauseRx);
    if(m){
	return [m[1],m[2]];
    }else{
	m=clause.match(constRx);
	if(m){
	    return [clause];
	}else{
	    return null;
	}
    }
}

// Takes "2d6+1d4-3" and returns 
// ['2d6','+','1d4','-','3']
// returns [] if not in the right form.
var termRx=/^([^+\-]+)([+\-]?)(.*)$/;
function parseTerms(line){
    var m=line.match(termRx);
    
    if(m){ // pulled off a term
	var car=m[1]; // first term
	var op=m[2]; // + or -
	var cdr=m[3]; // Rest of the string (if any)
	
	if(car && !op){ // last term
	    return [car];
	}else if(car && op && !cdr){ // bad form!
	    return [];
	}else if(car && op && cdr){ // recurse
	    var cdrTokens=parseTerms(cdr);
	    if(cdrTokens.length===0){ // bad form
		return []; // return empty
	    }else{
		var terms=[car,op];
		terms.push.apply(terms,cdrTokens);
		return terms;
	    }
	}
    }else{ // problem in the first term, fail
	return [];
    }
}

// '2d6+3' => 11
// returns null if the wrong form
function rollExpression(expr){
    var terms=parseTerms(expr);

    var lastOp='+'; // First term is positive
    var isTerm=true; // true for term, false for operator

    var sum=0;

    if(terms.length===0){
	return null;
    }

    for(var k=0;k<terms.length;k++){
	var token=terms[k];
	
	if(!isTerm){
	    isTerm=true;
	    lastOp=token;
	}else{
	    isTerm=false;
	    var clause=parseClause(token);

	    if(clause===null){ // invalid clause
		return null;
	    }

	    var val=roll.apply(null,clause);

	    if(lastOp==='+'){
		sum+=val;
	    }else if(lastOp==='-'){
		sum-=val;
	    }
	}
    }

    return sum;
}